/**************************************************************************
*   Copyright (C) 2008 by Michael Hoffer <info@michaelhoffer.de>          *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as               *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public             *
*   License along with this program; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "CbfReader.h"

std::string CbfReader::getName() const
{
	return std::string("CBF");
}

int CbfReader::readHeader( std::ifstream &cbfTapin )
{
	std::cout << getName() << "::read image:" << std::endl;

	int nx = 0;
	int ny = 0;


	//check if we can read the file and if it is the right file type
	int formatResult = checkImageFormat( cbfTapin );
	if ( formatResult != 0 ) return formatResult;

	// write the file header to a string
	std::string textInfo;
	int stream2StringResult = stream2String( cbfTapin, textInfo );
	if ( stream2StringResult != 0 ) return stream2StringResult;

	// find the beginning of the mime-type header
	std::string::size_type pos = textInfo.find( "_array_data.data" );

	if ( pos == std::string::npos )
	{
		std::cerr << "_array_data.data not found!" << std::endl;
		return -3; // Wrong image format
	}

	// filter comments in mime-type section
	std::string withoutComments;
	int filterCommentsResult = filterComments( textInfo, pos, withoutComments );

	if ( filterCommentsResult != 0 ) return filterCommentsResult;


	// store cif-header as array of lines
	std::vector<std::string> cifHeader = tokenize( withoutComments );

	// check if cif-header "open tags" are correct
	if ( cifHeader[1].find( ";" ) == std::string::npos )
	{
		std::cerr << ">> missing ';' after '_array_data.data'!" << std::endl;
		return -3; // Wrong image format
	}

	if ( cifHeader[2].find( "--CIF-BINARY-FORMAT-SECTION--" ) == std::string::npos )
	{
		std::cerr << ">> '--CIF-BINARY_FORMAT_SECTION--' not found!" << std::endl;
		return -3; // Wrong image format
	}

	int offset = 3;


	// tokenize cif-header entries using ':' as seperator
	std::vector<std::vector<std::string> > cifEntries;

	for ( unsigned int i = offset; i < cifHeader.size(); i++ )
	{
		std::vector<std::vector<std::string> > finalTokenList;
		std::vector<std::string> tmpTokenList = tokenize( cifHeader[i], ':' ) ;

		finalTokenList.push_back( std::vector<std::string>() );

		if ( tmpTokenList.size() == 1 )
		{

			// we assume only one '=' per line
			finalTokenList[0] =  tokenize( cifHeader[i], '=' ) ;
		}
		// we only assume one ':' per line
		else if ( tmpTokenList.size() == 2 )
		{

			finalTokenList[0].push_back( tmpTokenList[0] );

			tmpTokenList = tokenize( tmpTokenList[1], ';' ) ;

			for ( unsigned int j = 0; j < tmpTokenList.size();j++ )
			{
				// we only assume one '=' per line
				std::vector<std::string> tmpTokenList2 = tokenize( tmpTokenList[j], '=' ) ;

				if ( tmpTokenList2.size() == 1 )
				{
					finalTokenList[0].push_back( tmpTokenList2[0] );
				}
				else
				{
					finalTokenList.push_back( tmpTokenList2 );
				}

			}
		}

		// add tokens to cifEntry list
		for ( unsigned int j = 0; j < finalTokenList.size();j++ )
		{
			cifEntries.push_back( finalTokenList[j] );
		}
	}


	for ( unsigned int i = 0; i < cifEntries.size();i++ )
	{

		std::vector<std::string> tokenList = cifEntries[i];

		tokenList[0] = removeWhiteSpace( tokenList[0], ' ' );
		tokenList[0] = removeWhiteSpace( tokenList[0], '\t' );
		tokenList[0] = toUpper( tokenList[0] );

		int tokenID = getTokenID( tokenList[0], getTokenList() );
		//std::cout << "token (ID: " << tokenID  << "): " << tokenList[0]  << std::endl;

		bool argumentParsed = false;

		// ContentType
		if ( tokenID == 0 )
		{
			argumentParsed = findToken( "application/octet-stream", tokenList );

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: wrong content type!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> ContentType: OK!" << std::endl;
			}
		}

		// conversions
		if ( tokenID == 1 )
		{
			argumentParsed = findToken( "\"x-CBF_BYTE_OFFSET\"", tokenList );

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: wrong conversions!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> conversions: OK!" << std::endl;
			}
		}

		// Content-Transfer-Encoding
		if ( tokenID == 2 )
		{
			argumentParsed = findToken( "BINARY", tokenList );

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: wrong transfer encoding!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> transfer encoding: OK!" << std::endl;
			}
		}

		// X-Binary-Element-Type
		if ( tokenID == 3 )
		{
			argumentParsed =  tokenList.size() > 1;

			if ( argumentParsed )
			{
				bool signedToken = tokenList[1].find( "signed" ) < std::string::npos;
				bool bitToken =  tokenList[1].find( "32-bit" ) < std::string::npos;
				bool integerToken =  tokenList[1].find( "integer" ) < std::string::npos;

				argumentParsed = signedToken && bitToken && integerToken;
			}

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: wrong binary element type!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> binary element type: OK!" << std::endl;
			}
		}

		// X-Binary-Element-Byte-Order
		if ( tokenID == 4 )
		{
			argumentParsed = findToken( "LITTLE_ENDIAN", tokenList );

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: wrong element byte order!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> element byte order: OK!" << std::endl;
			}
		}

		// X-Binary-Size-Fastest-Dimension
		if ( tokenID == 5 )
		{
			argumentParsed =  tokenList.size() == 2;

			if ( argumentParsed )
			{
				tokenList[1] = removeWhiteSpace( tokenList[1], ' ' );
				tokenList[1] = removeWhiteSpace( tokenList[1], '\t' );

				tokenList[1] = removeWhiteSpace( tokenList[1], '\r' );
				tokenList[1] = removeWhiteSpace( tokenList[1], '\n' );

				nx = atoi( tokenList[1].c_str() );

				// check if we got a number
				argumentParsed = nx > 0;
			}

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: missing fastest dimension size!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> fastest dimension size: " << nx << " OK!" << std::endl;
			}
		}

		// X-Binary-Size-Second-Dimension
		if ( tokenID == 6 )
		{
			argumentParsed =  tokenList.size() == 2;

			if ( argumentParsed )
			{
				tokenList[1] = removeWhiteSpace( tokenList[1], ' ' );
				tokenList[1] = removeWhiteSpace( tokenList[1], '\t' );

				tokenList[1] = removeWhiteSpace( tokenList[1], '\r' );
				tokenList[1] = removeWhiteSpace( tokenList[1], '\n' );

				ny = atoi( tokenList[1].c_str() );

				// check if we got a number
				argumentParsed = ny > 0;
			}

			if ( !argumentParsed )
			{
				std::cerr << ">> Error: missing second dimension size!" << std::endl;
				return -3; // Wrong image format
			}
			else
			{
				argumentParsed = true;
				std::cout << ">> second dimension size: " << ny << " OK!" << std::endl;
			}
		}

	}

	_nx = nx;
	_ny = ny;

	return 0;
}

int CbfReader::readBinary( std::ifstream& tapin )
{
	int nx = _nx;
	int ny = _ny;

	char uc0, uc1;
	char sc;
	int diff, tmp0, tmp1, pixvalue = 0;
	long long int ldiff;

	for ( int i = 0;i < nx*ny;i++ )
	{
		if ( tapin.get( sc ) == 0 )return -4;
		diff = sc;
		if ( diff == -128 )
		{
			if ( tapin.get( uc0 ) == 0 )return -4;
			if ( tapin.get( uc1 ) == 0 )return -4;
			tmp0 = uc0;
			if ( uc0 < 0 ) tmp0 += 256; // converting signed to unsigned char
			diff = uc1;
			if ( diff < 0 ) diff += 256; // converting signed to unsigned char
			diff = tmp0 + 256 * diff;
			if ( diff > 32767 )diff -= 65536;
			if ( diff == -32768 )
			{
				if ( tapin.get( uc0 ) == 0 )return -4;
				if ( tapin.get( uc1 ) == 0 )return -4;
				tmp0 = uc0;
				if ( tmp0 < 0 ) tmp0 += 256; // converting signed to unsigned char
				tmp1 = uc1;
				if ( tmp1 < 0 ) tmp1 += 256; // converting signed to unsigned char
				diff = tmp0 + 256 * tmp1;
				if ( tapin.get( uc0 ) == 0 )return -4;
				if ( tapin.get( uc1 ) == 0 )return -4;
				tmp0 = uc0;
				if ( tmp0 < 0 ) tmp0 += 256; // converting signed to unsigned char
				tmp1 = uc1;
				if ( tmp1 < 0 ) tmp1 += 256; // converting signed to unsigned char
				ldiff = tmp0 + 256 * tmp1;
				ldiff = ldiff * 65536 + diff;
				if ( ldiff > 2147483647 )ldiff -= 4294967296LL;
				diff = ldiff; //overflow impossible for real images
			}
		}
		pixvalue += diff;
		_memory[i] = pixvalue;
	}

	return 0;
}

bool CbfReader::isMarker( std::ifstream &tapin )
{
	std::ifstream::pos_type streamPos = tapin.tellg();

	char binaryMarker[5] = {12, 26, 4, 213, '\0'};
	char buffer[5] = {0, 0, 0, 0, '\0'};

	bool equal = true;

	tapin.read( buffer, sizeof( char )*4 );

	if ( tapin.bad() ) return false;

	// compare
	for ( int i = 0; i < 4; i++ )
	{
		equal = ( buffer[i] == binaryMarker[i] );
		if ( !equal ) break;
	}

	if ( equal == false )
	{
		tapin.seekg( streamPos );
	}

	return equal;
}


std::vector<std::string> CbfReader::getTokenList()
{
	std::vector<std::string> token;
	token.push_back( "Content-Type" );			// case 0
	token.push_back( "conversions" );			// case 1
	token.push_back( "CONTENT-TRANSFER-ENCODING" );		// case 2
	token.push_back( "X-BINARY-ELEMENT-TYPE" );		// case 3
	token.push_back( "X-BINARY-ELEMENT-BYTE-ORDER" );	// case 4
	token.push_back( "X-BINARY-SIZE-FASTEST-DIMENSION" );	// case 5
	token.push_back( "X-BINARY-SIZE-SECOND-DIMENSION" );	// case 6
	//token.push_back( "x-BINARY-SIZE-PADDING" );
	//token.push_back( "CONTENT-MD5" );
	//token.push_back( "X-BINARY-NUMBER-OF-ELEMENTS" );
	//token.push_back( "X-BINARY-SIZE" );
	//token.push_back( "X-BINARY-ID" );

	return token;
}

int  CbfReader::filterComments( std::string in, int index, std::string &out )
{
	int result = 0;
	int commentLevel = 0;

	for ( unsigned int i = index; i < in.size();i++ )
	{

		if ( in[i] == '(' )
		{
			commentLevel++;
		}
		if ( in[i] == ')' )
		{
			commentLevel--;
		}

		if ( commentLevel == 0 )
		{
			if ( in[i] != '(' && in[i] != ')' )
			{
				out += in[i];
			}
		}
	}

	if ( commentLevel != 0 )
	{
		result = -3;  // Wrong image format
		std::string error;
		if ( commentLevel < 0 )
		{
			error = "too many ')'.";
		}
		else
		{
			error = "missing ')'.";
		}
		std::cerr << ">> Syntax error in file header: " << error << std::endl;
	}
	return result;
}

int CbfReader::checkImageFormat( std::ifstream& tapin )
{
	if ( tapin == 0 )return -2; // Cannot open image file

	{ // Check image file format
		std::string s( "###CBF: " );
		char label[9];
		for ( int i = 0;i < 8;i++ )
			if ( tapin.get( label[i] ) == false )return -4; // Cannot read image file
		label[8] = '\0';
		std::string sLabel = ( std::string )label;

		//sLabel = toUpper(sLabel);std::string result = s;

		// convert sLabel to upper case
		for ( unsigned int i = 0; i < sLabel.length(); i++ )
		{
			sLabel[i] = toupper( sLabel[i] );
		}

		if ( sLabel != s )return -3; // Wrong image format

	}

	return 0;
}

