/**************************************************************************
*   Copyright (C) 2008 by Michael Hoffer <info@michaelhoffer.de>          *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as               *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public             *
*   License along with this program; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "ImageArray.h"
#include <assert.h>

#include <limits.h>
#include <string>
#include <sstream>
#include <vector>
#include <iostream>
#include <algorithm>
#include <cctype>
#include <fstream>
#include <cmath>

#include <QString>
#include <QRegExp>
#include <QMessageBox>

#include <stdio.h>

ImageArray::ImageArray()
{
	_memory = NULL;
	_nCols = 0;
	_nRows = 0;

	_limitsOutOfDate = true;
	_minValue = INT_MIN;
	_maxValue = INT_MAX;

	_step = 1;
	_slope = 100;
	_percentile = 0;
	_customColorMap = false;

	_imageReaders = std::vector<ImageReader*>();
}

ImageArray::ImageArray( const ImageArray& imageArray )
{

	std::cout << "ImageArray::ImageArray(const ImageArray& imageArray)" << std::endl;

	_nCols = imageArray.nCols();
	_nRows = imageArray.nRows();

	if ( imageArray.isInitialized() )
	{
		_memory = new int[nCols()*nRows()];
	}
	else
	{
		_memory = NULL;
	}


	for ( unsigned int i = 0; i < nCols()*nRows();i++ )
	{
		_memory[i] = imageArray._memory[i];
	}

	_limitsOutOfDate = imageArray._limitsOutOfDate;
	_minValue = imageArray._minValue;
	_maxValue = imageArray._maxValue;

	_step = imageArray._step;
	_slope = imageArray._slope;
	_percentile = 0;
	_customColorMap = false;


	_imageReaders = std::vector<ImageReader*>();
}

ImageArray::ImageArray( unsigned int nCols, unsigned int nRows )
{
	_memory = new int[ nCols * nRows ];

	_nCols = nCols;
	_nRows = nRows;

	zero();

	_limitsOutOfDate = true;
	_minValue = INT_MIN;
	_maxValue = INT_MAX;
	_percentile = 0;
	_slope = 100;
	_customColorMap = false;

	_imageReaders = std::vector<ImageReader*>();
}

ImageArray::ImageArray( std::string fileName )
{

	_memory = NULL;
	_nCols = 0;
	_nRows = 0;

	_limitsOutOfDate = true;
	_minValue = INT_MIN;
	_maxValue = INT_MAX;

	_step = 1;
	_slope = 100;
	_percentile = 0;
	_customColorMap = false;

	_imageReaders = std::vector<ImageReader*>();
	_imageReaders.push_back( new CbfReader() );

	bool imageLoaded = false;

	for ( unsigned int i = 0; i < _imageReaders.size();i++ )
	{
		std::cout << "<checking format=\""  << _imageReaders[i]->getName()
		<< "\">\n";

		// create file stream
		std::ifstream tapin( fileName.c_str(), std::ios_base::binary );

		int ier = _imageReaders[i]->readHeader( tapin );

		if ( ier == 0 )
		{

			_nCols = _imageReaders[i]->getNX();
			_nRows = _imageReaders[i]->getNY();

			_memory = new int[nRows()*nCols()];

			_imageReaders[i]->setMemory( _memory );

			_limitsOutOfDate = true;
			_minValue = INT_MIN;
			_maxValue = INT_MAX;

			_step = 1;
			_percentile = 0;

			ier = _imageReaders[i]->readBinary( tapin );

		}

		std::string errorMessage = "";

		switch ( ier )
		{
		case  0:
			//std::cerr << "Image successfully read" << std::endl;
			imageLoaded = true;
			break;
		case - 2:
			errorMessage = "Cannot open image file";
			break;
		case - 3:
			errorMessage = "Wrong image file format";
			break;
		case - 4:
			errorMessage = "Cannot read image file";
			break;
		default:
			errorMessage = "Programming error";
		}

		if ( ier == 0 )
		{
			std::cout << "</checking>\n";
			break;
		}
		else
		{
			std::cerr << ">> Error: " << errorMessage << std::endl;

			QMessageBox::critical( NULL, QString( "XDS-Viewer" ),
			                       errorMessage.c_str() );
		}
	}

	//-----------------------------------------------------------------------------

	if ( !imageLoaded )
	{
		std::cout << ">> Image Format not recognized!" << std::endl;
		std::cout << "</checking>\n";
	}

	computeLimits();
}

const bool ImageArray::isInitialized() const
{
	return _memory != NULL;
}

int ImageArray::pix_percentile( float p )
{
	// returns the p-th percentile value of the image pixels
	int npix = nCols() * nRows();

	int pixel_min = INT_MAX;
	for ( int i = 0;i < npix;i++ )pixel_min = ( pixel_min > _memory[i] ) ? _memory[i] : pixel_min;

	double pixel_mean = 0.0;
	for ( int i = 0;i < npix;i++ )pixel_mean += ( double )_memory[i];
	if ( npix > 1 )pixel_mean /= ( double )npix;

	const int NHIST = 10000;
	int histogram[NHIST];
	for ( int i = 0;i < NHIST;i++ )histogram[i] = 0;

	double f = 1.0;
	{
		double d = 1.0 + 2.0 * ( pixel_mean - ( double )pixel_min );
		if ( d > ( double )NHIST )f = (( double )NHIST ) / d;
	}

	for ( int k = 0, i = 0;i < npix;i++ )
	{
		k = ( int )( f * ( double )( _memory[i] - pixel_min ) );
		if ( k < 0 )k = 0;
		if ( k >= NHIST )k = NHIST - 1;
		histogram[k]++;
	}

	double h = 0.0;
	for ( int npix1 = ( int )( p * (( float )npix ) / 100.0 ), n = 0, k = 0;k < NHIST;k++ )
	{
		n += histogram[k];
		h = ( double )k;
		if ( n > npix1 )break;
	}

	return int( h / f ) + pixel_min;
}



const unsigned int ImageArray::nCols() const
{
	return _nCols;
}


const unsigned int ImageArray::nRows() const
{
	return _nRows;
}


void ImageArray::zero()
{
	// Fills memory with zero-bytes.
	for ( unsigned int i = 0; i < nCols()*nRows();i++ )
	{
		_memory[i] = 0;
	}
}
ImageArray& ImageArray::operator= ( const ImageArray & imageArray )
{

	std::cout << "ImageArray::operator="  << std::endl;
	if ( &imageArray == this )
	{
		// nothing to do
	}
	else
	{
		_nCols = imageArray.nCols();
		_nRows = imageArray.nRows();

		if ( _memory != NULL )
		{
			delete[] _memory;
		}

		if ( imageArray.isInitialized() )
		{
			_memory = new int[_nCols*_nRows];
		}
		else
		{
			_memory = NULL;
		}

		// copy content
		for ( unsigned int i = 0; i < _nCols*_nRows; i++ )
		{
			_memory[i] = imageArray._memory[i];
		}

		_limitsOutOfDate = imageArray._limitsOutOfDate;
		_minValue = imageArray._minValue;
		_maxValue = imageArray._maxValue;

		_step = imageArray._step;
		_slope = imageArray._slope;
		_customColorMap = false;

	}
	return *this;
}

const int& ImageArray::operator()( unsigned int x, unsigned int y ) const
{
	assert( x >= 0 );
	assert( x < nCols() );

	assert( y >= 0 );
	assert( y < nRows() );

	return _memory[x+nCols()*y];
}


int& ImageArray::operator()( unsigned int x, unsigned int y )
{
	assert( x >= 0 );
	assert( x < nCols() );

	assert( y >= 0 );
	assert( y < nRows() );

	//_limitsOutOfDate = true;

	return _memory[x+nCols()*y];
}

QImage ImageArray::toQImage()
{
	QImage image( nCols(), nRows(), QImage::Format_Indexed8 );

	image.setColorTable( ColorTable() );

	_limitsOutOfDate = true;
	computeLimits();

	computeColorMap();

	// set pixels
	for ( unsigned int y = 0; y < nRows();y++ )
	{
		for ( unsigned int x = 0; x < nCols(); x++ )
		{
			image.setPixel( x, y, getColorIndexValue(( *this )( x, y ) ) );
		}
	}

	return image;
}

void ImageArray::setCustomColorMap(ImageArray &img){
	_customColorMap = true;
	_step = img._step;
}

const int ImageArray::getSlope() const{
	return _slope;
}

void ImageArray::setSlope(int slope){
	_slope = slope;
}

void ImageArray::computeColorMap()
{
	if (!_customColorMap){
		int length = abs( _percentile - getMinValue() );

		if ( length > 0 )
		{
			_step = (double)_slope / (double)length; // _slope is usually 100
		}
	}
}
/**
 * Each pixel value is assigned to an index value between 0 and 255
 */
const unsigned int ImageArray::getColorIndexValue( const int i )
{
	double value = double( i -getMinValue() ) * _step;

	//if ( value > 100 )
	//{
	//        value = 100.0 + 154.0 * log(i-_percentile + 1)/ (log(getMaxValue()-_percentile+1)+1.0);
	//}

	//if ( value > 255 ) value = 255;

	//double value = 100.0 * log(i - getMinValue() +1) / log(_percentile - getMinValue() + 1);

	//double value = 100.0 * sqrt(i - getMinValue() +1) / sqrt(_percentile - getMinValue() + 1);


	if ( value > 255 ) value = 255;

	return ( unsigned int )( value );
}

const int ImageArray::getMinValue()
{
	computeLimits();
	return _minValue;
}

const int ImageArray::getMaxValue()
{
	computeLimits();
	return _maxValue;
}

void ImageArray::computeLimits()
{
	if ( _limitsOutOfDate )
	{
		_minValue = INT_MAX;
		_maxValue = INT_MIN;

		for ( unsigned int i = 0; i < nCols()*nRows();i++ )
		{
			if ( _memory[i] < _minValue )
			{
				_minValue = _memory[i];
			}

			if ( _memory[i] > _maxValue )
			{
				_maxValue = _memory[i];
			}
		}

		_percentile = pix_percentile( 75.0 );

		_limitsOutOfDate = false;
	}
}
ImageArray::~ImageArray()
{
	if ( _memory != NULL )
	{
		delete[] _memory;
		_memory = NULL;
	}

	for ( unsigned int i = 0; i < _imageReaders.size();i++ )
	{
		if ( _imageReaders[i] != NULL )
		{
			delete _imageReaders[i];
		}
	}
}
