/**************************************************************************
*   Copyright (C) 2008 by Michael Hoffer <info@michaelhoffer.de>          *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as               *
*   published by the Free Software Foundation; either version 2 of the    *
*   License, or (at your option) any later version.                       *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public             *
*   License along with this program; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

#include "mainwindow.h"
#include <QtGui>
#include <QProcess>
#include <QMessageBox>

#include <iostream>
#include <unistd.h>
#include <sstream>

#include<StringTools.h>

MainWindow::MainWindow( QWidget *parent ): QMainWindow( parent )
{
	setupUi( this );

	canvas->setListener( this );

	colorTable = ColorTable();

	_title = "XDS-Viewer";

	setWindowTitle( _title );

	slopeEdit->setValidator( new QIntValidator( 1, 1000, this ) );

	_helpUsageWindow = new HelpWindow( ":/doc/help_usage.html", this );
	_helpAboutWindow = new HelpWindow( ":/doc/help_about.html", this );

}

void MainWindow::valueChanged( int value, int x, int y )
{
	QString string = "x=" + QString::number( x ) + ", y=" +
	                 QString::number( y ) + "\n value=" + QString::number( value );
	textDisplay->setText( string );

}

void MainWindow::loadCbfFile( std::string fileName )
{
	imageData = ImageArray( fileName );
	
	if ( imageData.isInitialized() )
	{
		canvas->setImageArray( imageData );
		on_pbSlope_clicked(); // update slope value
		on_colorSlider_valueChanged();
		canvas->repaint();

		int min = canvas->getImageArray().getMinValue();
		int max = canvas->getImageArray().getMaxValue();

		setWindowTitle( _title + " (" + fileName.c_str() + ")" );
	}
}

void MainWindow::loadFile( std::string fileName )
{
	QString tmpStr = fileName.c_str();
	bool isCbf = tmpStr.endsWith( ".CBF", Qt::CaseInsensitive );

	bool fileChosen = fileName.size() > 0;


	if ( fileChosen )
	{
		if ( isCbf )
		{
			loadCbfFile( fileName );
		}
		else
		{
			int pid = getpid();
			std::stringstream tmpFileNameStream;
			tmpFileNameStream << "/tmp/xds_viewer_tmp_img_" << pid;
			QString tmpFileName = tmpFileNameStream.str().c_str();
			QString inputFileName = fileName.c_str();

			QProcess toCbf;

			// display system path (debug only)
			// QMessageBox::information( NULL, QString( "XDS-Viewer" ), toCbf.systemEnvironment().join(":"));
                                         
			toCbf.start( "2cbf" );

			if ( !toCbf.waitForStarted() )
			{
				// error dialog
				QMessageBox::critical( NULL, QString( "XDS-Viewer" ),
				                       QString( "Cannot open file! For image formats other than \".cbf\" you" )
				                       + QString( " need to install the \"2cbf\" script. Please make sure you" )
				                       + QString( " have added it to the executable path." ) );
				return;

			}

			QString arguments = inputFileName + " " + tmpFileName;

			toCbf.write( arguments.toStdString().c_str() );
			toCbf.closeWriteChannel();


			if ( !toCbf.waitForFinished() )
			{
				// TODO error message
				return;
			}

			QByteArray toCbfOutput = toCbf.readAll();

			if ( toCbfOutput.size() > 0 )
			{
				QMessageBox::critical( NULL, QString( "XDS-Viewer" ),
				                       toCbfOutput );
				return;
			}


			loadCbfFile( tmpFileName.toStdString() );

			setWindowTitle( fileName.c_str() );

			QProcess rmTmpFile;
			rmTmpFile.start( "rm", QStringList() << " -f " << tmpFileName );

			if ( !rmTmpFile.waitForFinished() )
			{
				// error dialog
				// cannot remove tmp file
				QMessageBox::critical( NULL, QString( "XDS-Viewer" ),
				                       "Caannot delete obsolete tempfile: " + tmpFileName );

			}
		}
	}

}


void MainWindow::startPainting()
{
	//this->repaint();
}

void MainWindow::on_pbSlope_clicked()
{
	int slope = slopeEdit->text().toInt();
	
	if (canvas->isInitialized()){
		canvas->getImageArray().setSlope( slope );
		canvas->updateImageArray();
	}
	
	
	on_colorSlider_valueChanged();
	canvas->repaint();
}

void MainWindow::on_slopeEdit_returnPressed(){
	on_pbSlope_clicked();
}

void MainWindow::on_actionQuit_triggered()
{
	this->close();
}

void MainWindow::on_actionLoadImage_triggered()
{

	fileName = QFileDialog::getOpenFileName( this,
	           tr( "Open Image" ), fileName.c_str(),
	           tr( "Image Files (*)" ) ).toStdString();
	loadFile( fileName );
}


void MainWindow::on_colorSlider_valueChanged()
{
	colorTable.generate( colorSlider->value() );

	canvas->setColorTable( colorTable );
	canvas->updateImageBuffer();
	canvas->repaint();

	colorCanvas->setColorTable( colorTable );
	colorCanvas->repaint();
}

void MainWindow::on_actionUsage_triggered(){
	_helpUsageWindow->show();	
}

void MainWindow::on_actionAbout_triggered(){
	_helpAboutWindow->show();
}
